
// linux-like acyclic list (DLL with NULL ends)
// uses pointer arithmetic to get list node data

#include <stdlib.h>             // malloc, free, NULL
#include <stddef.h>             // offsetof

/**
 * list_entry - get the pointer to data struct for this entry
 * @ptr:        the &struct list_link pointer.
 * @type:       the type of the data struct the link is embedded in.
 * @member:     the name of the link field within the struct.
 */
#define list_entry(ptr, type, member) ({                        \
        const typeof( ((type *)0)->member ) *__mptr = (ptr);    \
        (type *)( (char *)__mptr - offsetof(type,member) );})

// link to embed in any struct with data
struct list_link {
    struct list_link *next;
    struct list_link *prev;
};

// data struct
struct Node {
    int d;
    struct list_link link;
    char c;
};

int main()
{
    struct list_link *h = NULL;
    struct list_link *t = NULL;
    struct list_link *p;
    struct Node *node;
    int i;
    for (i = 0; i < 10; ++i) {
        node = (struct Node *) malloc(sizeof(struct Node));
        p = &(node->link);
        if (NULL == h) {
        p->next = NULL;
            p->prev = NULL;
            h = p;
        } else {
        p->next = NULL;
            p->prev = t;
            t->next = p;
        }
        t = p;
        node->d = 10;
    }
    for (p = h; p != NULL; p = p->next) {
        node = list_entry(p, struct Node, link);
        node->d = 20;
    }
    while (h != NULL) {
        node = list_entry(h, struct Node, link);
        p = h->next;
        if (p == NULL)
            h = t = NULL;
        else {
            h = p;
            p->prev = NULL;
        }
        free((void *) node);
    }
    return h != NULL;
}
